import yaml
import marshmallow_dataclass

from fnmatch import fnmatch
from typing import Dict, Set, List, Any, Tuple, Optional
from logging import Logger, Filter, getLogger
from pluggy import HookimplMarker  # type: ignore

from dataclasses import dataclass
from marshmallow.exceptions import MarshmallowError, ValidationError

ENCAB = "encab"
LOG_SANITIZER = "log_sanitizer"

mylogger = getLogger(LOG_SANITIZER)


class ConfigError(ValueError):
    pass


@dataclass
class LogSanitizerSettings(object):
    """
    the log sanitizer settings

    eample:

    .. code-block:: yaml

        encab:
            debug: true
            halt_on_exit: False
        extensions:
            log_sanitizer:
                enabled: true
                settings:
                    override: false
                    patterns:
                        - "*MAGIC*"

    This class contains the extensions/log_sanitizer/settings content.
    """

    patterns: Optional[List[str]]
    """set sensitive environmen variable name patterns whose values will be masked.
       UNIX file pattern rules are used (see https://docs.python.org/3/library/fnmatch.html#module-fnmatch)       
    
       example:
       
        ``*MAGIC*`` -- all variable names containing the string MAGIC 
    """

    override: Optional[bool]
    """if True, builtin patterns are overridden """

    def __post_init__(self):
        self.patterns = self.patterns or list()

        self.override = False if self.override is None else self.override

    @staticmethod
    def load(settings: Dict[str, Any]) -> "LogSanitizerSettings":
        try:
            ConfigSchema = marshmallow_dataclass.class_schema(LogSanitizerSettings)
            return ConfigSchema().load(settings)  # type: ignore
        except ValidationError as e:
            msg = e.args[0]
            if isinstance(msg, dict):
                msg = yaml.dump(msg, default_flow_style=False)

            raise ConfigError(f"\n\n{LOG_SANITIZER}:\n{msg}")
        except MarshmallowError as e:
            raise ConfigError(e.args)


class SanitizingFilter(Filter):
    def __init__(self, sensitive_strings: Set[str]) -> None:
        super().__init__()
        self.sensitive_strings = sensitive_strings

    def sanitize(self, value: Any) -> Any:
        for s in self.sensitive_strings:
            if isinstance(value, int) or isinstance(value, float):
                value = -1 if s == str(value) else value
            else:
                value = str(value).replace(s, "*" * len(s))
        return value

    def sanitize_all(self, args: Tuple[Any]):
        return tuple([self.sanitize(arg) for arg in args])

    def filter(self, record):
        record.msg = self.sanitize(record.msg)
        assert isinstance(record.args, tuple)
        record.args = self.sanitize_all(record.args)
        return True


extension_impl = HookimplMarker(ENCAB)


class LogSanitizerExtension(object):
    """
    Simple log sanitizer for environment variable values.
    It looks for values of environment variables whose names match
    certain patterns and replaces them with asterics.

    Predefined patterns are:

    - ``*KEY*``
    - ``*SECRET*``
    - ``*PASSWORD*``

    For patterns, UNIX file pattern rules are used (see https://docs.python.org/3/library/fnmatch.html#module-fnmatch)
    They can be extended or overriden in the extension settings.

    Example:

    Suppose you have the variable MY_PASSWORD set like this:

    MY_PASSWORD=s3cR37

    The name MY_PASSWORD matches the predifined pattern ``*PASSWORD*``,
    so if you run a program like ``echo $MY_PASSWORD``,
    the output will be raplaced by ``******``.

    This extension is enabled by default and can be disabled in the extension settings.

    """

    PATTERNS = ["*KEY*", "*SECRET*", "*PASSWORD"]

    def __init__(self) -> None:
        self.sensitive_strings: Set[str] = set()
        self.settings = LogSanitizerSettings(patterns=self.PATTERNS, override=False)
        self.enabled = True

    @extension_impl
    def validate_extension(self, name: str, enabled: bool, settings: Dict[str, Any]):
        if name == LOG_SANITIZER:
            LogSanitizerSettings.load(settings)
            mylogger.info("settings are valid.", extra={"program": ENCAB})

    @extension_impl
    def configure_extension(self, name: str, enabled: bool, settings: Dict[str, Any]):
        if name != LOG_SANITIZER:
            return

        if not enabled:
            self.enabled = False
            return

        self.settings = LogSanitizerSettings.load(settings)

        if not self.settings.override:
            assert isinstance(self.settings.patterns, list)
            patterns = self.settings.patterns
            self.settings.patterns = [*patterns, *self.PATTERNS]

    def is_sensitive(self, name: str) -> bool:
        assert isinstance(self.settings.patterns, list)
        patterns = self.settings.patterns
        for pattern in patterns:
            if fnmatch(name.upper(), pattern.upper()):
                return True
        return False

    @extension_impl
    def extend_environment(self, program_name: str, environment: Dict[str, str]):
        if not self.enabled:
            return

        for name, value in environment.items():
            if self.is_sensitive(name):
                self.sensitive_strings.add(value)

    @extension_impl
    def update_logger(self, program_name: str, logger: Logger):
        if program_name == ENCAB:
            if self.enabled:
                patterns = str(self.settings.patterns)
                mylogger.debug(
                    "patterns: %s",
                    patterns,
                    extra={"program": program_name},
                )
            else:
                mylogger.info(
                    "log_sanitizer is disabled", extra={"program": program_name}
                )

        if not self.enabled:
            return

        logger.addFilter(SanitizingFilter(self.sensitive_strings))



from logging import Logger
from typing import Dict, Any, List
from importlib import import_module
from pluggy import HookspecMarker, PluginManager, PluginValidationError  # type: ignore


ENCAB = "encab"

extension_method = HookspecMarker(ENCAB)


class Extensions(object):
    """
    Represents the connection point between encab and its extensions

    Extension methods specified here are implemented by extensions

    example extension code:

    .. code-block:: python

        from pluggy import HookimplMarker  # type: ignore

        ENCAB = "encab"
        extension_impl = HookimplMarker(ENCAB)

        @extension_impl
        def validate_extension(
            self,
            name: str,
            enabled: bool,
            config: Dict[str, Any]):
            '''
                your implementation
            '''
            pass

    """

    def __init__(
        self,
    ) -> None:
        self.plugin_manager = PluginManager(ENCAB)
        self.plugin_manager.add_hookspecs(self)
        self.hook: Any = self.plugin_manager.hook

    @extension_method
    def validate_extension(self, name: str, enabled: bool, settings: Dict[str, Any]):
        """
        similar to :meth:`encab.extensions.Extensions.configure_extension`
        but the extension validates it's configuration during a dry run,
        where the extension isn't actually startet.

        see :meth:`encab.extensions.Extensions.configure_extension` for details.

        :param name: the extension name
        :type name: str
        :param enabled: if True: the extension is enabled
        :type enabled: bool
        :param settings: the extension settings from the encab config
        :type settings: Dict[str, Any]
        """
        self.hook.validate_extension(name=name, enabled=enabled, settings=settings)

    @extension_method
    def configure_extension(self, name: str, enabled: bool, settings: Dict[str, Any]):
        """
        configure_extension is called on startup for each extension.
        The extension picks up its own settings from the parameter ``config``
        by selecting for the name.

        Depening on the flag ``enabled``, each extension enables or disables its service.

        Encab config example:

        .. code-block:: yaml

            encab:
                debug: true
                halt_on_exit: False
            extensions:
                my_extension:
                    enabled: true
                    settings:
                        foo: bar


        The path for the settings is ``/extensions/<name>/settings``
        In this example, ``name``: would be ``my_extension`` and
        ``settings`` would be ``{'foo':'bar'}``.

        The extension will validate its settings and throw a ValueError (or a derivative thereof)
        with a descriptive error message if extension settings are invalid.

        :param name: the extension name
        :type name: str
        :param enabled: if True: the extension is enabled
        :type enabled: bool
        :param settings: the extension settings from the encab config
        :type settings: Dict[str, Any]
        """
        self.hook.configure_extension(name=name, enabled=enabled, settings=settings)

    @extension_method
    def extend_environment(self, program_name: str, environment: Dict[str, str]):
        """
        extend_environment is called whenever the environment of encab or a program
        (indicated by ``program_name``) is extended.

        The extension is free to alter the environment.

        :param program_name: the program name indicating the program the environment belongs to
        :type program_name: str
        :param environment: the environment as mutable dictionary
        :type environment: Dict[str, str]
        """
        self.hook.extend_environment(program_name=program_name, environment=environment)

    @extension_method
    def update_logger(self, program_name: str, logger: Logger):
        """
        update_logger is called whenever a logger for a program is introduced.

        The extension is free to alter the logger, e.g. by adding custom handlers.

        :param program_name: the logger's program name
        :type program_name: str
        :param logger: the newly introduced logger
        :type logger: Logger
        """
        self.hook.update_logger(program_name=program_name, logger=logger)

    def register(self, extensions: List[Any]) -> None:
        """
        register new extension as object or module

        :param extensions: a list of extension objects and/or modules
        :type extensions: List[Any]
        """
        for extension in extensions:
            self.plugin_manager.register(extension)

    def register_module(self, module_name: str):
        try:
            module = import_module(module_name)
            self.plugin_manager.register(module)
        except ModuleNotFoundError:
            raise FileNotFoundError(f"Extension module {module_name} not found")
        except ImportError as e:
            raise IOError(f"Failed to load extension module {module_name}: {str(e)}")
        except PluginValidationError as e:
            raise IOError(f"Failed to load extension module {module_name}: {str(e)}")


extensions = Extensions()
"""The singleton instance that's used by encab to call extension methods"""



import re
import io
import os
import shlex
import yaml
import marshmallow_dataclass

from yaml.error import YAMLError
from typing import Dict, Optional, Union, List, Any
from dataclasses import dataclass, fields
from marshmallow.exceptions import MarshmallowError, ValidationError
from logging import DEBUG, INFO
from abc import ABC

from .common.process import getUserId, getGroupId


class ConfigError(ValueError):
    """
    Configuration error
    """

    pass


@dataclass
class AbstractConfig(ABC):
    """
    Configuration base dataclass
    """

    @classmethod
    def create(cls, **args):
        """
        creates an instance and sets all missing optional fields to None

        :return: the data class
        :rtype: same as cls
        """
        config_class = dataclass(cls)  # type: ignore
        config_fields = fields(config_class)
        all_args = dict(args)
        for field in config_fields:
            name = field.name
            if name not in args:
                all_args[name] = None

        return config_class(**all_args)


@dataclass
class AbstractProgramConfig(AbstractConfig):
    """
    Common config for encab and programs section
    """

    environment: Optional[Dict[str, str]]
    """set additional environmment variables"""

    debug: Optional[bool]
    """if True, encab logs debug information. Default: False"""

    loglevel: Optional[Union[str, int]]
    """the log level. One of "CRITICAL", "FATAL", "ERROR", "WARN", "WARNING", "INFO", "DEBUG"."""

    umask: Optional[Union[str, int]]
    """the umask, either as octal string (e.g. "077") or integer"""

    user: Optional[Union[str, int]]
    """the user id or user name"""

    group: Optional[Union[str, int]]
    """the group id or group name"""

    join_time: Optional[float]
    """ 
    The join time is the time in seconds encab waits for a program to start/shutdown before 
    it continues with the next. Default: 1 seconds
    """

    def _set_umask(self):
        """
        sets umask and turns it into an int if necessary

        :raises ConfigError: is umask is given as string and is not an octal string
        """
        umask = self.umask

        if umask:
            if isinstance(umask, int):
                self.umask = int(umask)
            else:
                try:
                    self.umask = int(umask, 8)
                except ValueError:
                    raise ConfigError(
                        f"Expected octal string for umask but got: {umask}"
                    )
        else:
            self.umask = -1

    def set_user(self):
        """
        sets the user that runs the program

        steps:
        - checks wether a user is set.
        - determines the UID if the user is given as name.
        - checks if encab is run as root if a program is set to run with a different user.

        :raises ConfigError: if the user is unknown or encab is not run as root when needed
        """
        user = self.user

        if user:
            if isinstance(user, int) or user.isnumeric():
                self.user = int(user)
            else:
                try:
                    self.user = getUserId(user)
                except KeyError:
                    raise ConfigError(f"Unknown user {user}")

            if user != os.getuid() and os.getuid() != 0:
                raise ConfigError(
                    "Encab has to run as root to run it or programs as different user"
                )

    def set_group(self):
        """
        sets the group that runs the program

        steps:
        - checks wether the group is set.
        - determines the GID if the user is given as name.

        :raises ConfigError: if the user is unknown or encab is not run as root when needed
        """
        group = self.group

        if group:
            if isinstance(group, int) or group.isnumeric():
                self.group = int(group)
            else:
                try:
                    self.group = getGroupId(group)
                except KeyError:
                    raise ConfigError(f"Unknown group {group}")

    def _set_log_level(self):
        """
        selts the log level and turns it into an int if necessary

        :raises ConfigError: if an unspecified log level is given
        """
        levels = ["CRITICAL", "FATAL", "ERROR", "WARN", "WARNING", "INFO", "DEBUG"]

        level = self.loglevel

        if level and isinstance(level, str) and level not in levels:
            levels_printed = ", ".join(levels)
            raise ConfigError(
                f"Unsupported log level {level}. Supported levels are: {levels_printed}"
            )

        self.loglevel = DEBUG if self.debug else (level or INFO)

    def _set_environment(self):
        """
        sets the environment as dictionary

        :raises ConfigError: if variable names don't comply with POSIX 3.231 Name standard
        """
        self.environment = self.environment or dict()

        if not self.environment:
            return

        pattern = re.compile(r"^[a-zA-Z_]+[a-zA-Z0-9_]*")
        for name in self.environment.keys():
            if not pattern.match(name):
                raise ConfigError(
                    "Expected valid environment variable name (see POSIX 3.231 Name)"
                    f" but was '{name}'."
                )

    def __post_init__(self):
        """
        validates fields and sets default values

        :raises ConfigError: if fields have invalid values
        """

        self._unsetFields = [
            f.name for f in fields(self) if getattr(self, f.name) is None
        ]
        """name of fields which were not set in this configuration"""

        self._set_environment()
        self._set_log_level()
        self._set_umask()

        self.join_time = self.join_time or 1.0

    def was_unset(self, field_name: str) -> bool:
        """
        returns true if a field was initially unset

        :param field_name: the field name
        :type field_name: str
        :return: True: the field was initially unset, False: the field was set
        :rtype: bool
        """
        return field_name in self._unsetFields

    def extend(self, other: "AbstractProgramConfig"):
        """
        Extends configration with values from a given configuration

        If a field of the current configuration was not set, it is set by the field
        value of the given configuration.

        :param other: the given configuration
        :type other: AbstractProgramConfig
        """
        for name in self._unsetFields:
            if hasattr(other, name):
                v = getattr(other, name)
                if v is not None:
                    setattr(self, name, v)


@dataclass
class EncabConfig(AbstractProgramConfig):
    """
    reprensents the encab base configuration

    contains all static settings to start a program
    """

    halt_on_exit: Optional[bool]
    """halt on exit: if True, encab is halted after the main program ends. Default: False"""

    logformat: Optional[str]
    """
    Custom log format (see logrecord-attributes_). 
    The attribute "program" contains the ptogram name. 
    
    .. _logrecord-attributes: https://docs.python.org/3/library/logging.html#logrecord-attributes
    
    Default: ``%(levelname)-5.5s %(program)s: %(message)s``
    
    
    """

    dry_run: Optional[bool]
    """True: the configuration is checked but no program is started. Default: False"""

    def _set_log_format(self):
        default_format = "%(levelname)-5.5s %(program)s: %(message)s"
        debug_format = "%(asctime)s %(levelname)-5.5s %(module)s %(program)s %(threadName)s: %(message)s"

        if self.logformat:
            return

        if self.debug:
            self.logformat = debug_format
        else:
            self.logformat = default_format

    def __post_init__(self):
        super().__post_init__()

        self.dry_run = self.dry_run or False
        self.halt_on_exit = self.halt_on_exit or False

        self._set_log_format()


@dataclass
class ProgramConfig(AbstractProgramConfig):
    """
    reprensents a program configuraion

    contains all static settings to start a program
    """

    command: Union[str, List[str], None]
    """the command to be execution as list in POSIX style
        examples:
        
        .. code-block:: yaml
        
            program:
                main:
                    command:
                        echo "Test"
       
        .. code-block:: yaml
        
            program:
                main:
                    command:     
                        - echo 
                        - Test
    """

    sh: Union[str, List[str], None]
    """the shell script as string or list
    
        examples:
        
            .. code-block:: yaml
        
                program:
                    sh:
                        echo "Test"
                    
            .. code-block:: yaml
       
                program:
                    sh:     
                        - echo "Test1" 
                        - echo "Test2"
    """
    startup_delay: Optional[float]
    """The startup delay for this program in seconds. Default: 0 seconds"""

    directory: Optional[str]
    """the directory where the program is executed"""

    def __post_init__(self):
        """
        validates fields and sets default values

        :raises ConfigError: if fields have invalid values
        """
        super().__post_init__()

        sh = self.sh
        command = self.command

        if sh and command:
            raise ConfigError(
                "Please specify either sh or command attribute for programs"
            )

        if command:
            self.command = shlex.split(command) if isinstance(command, str) else command

        if sh:
            self.sh = sh if isinstance(sh, str) else "; ".join(sh)

        self.startup_delay = self.startup_delay or 0


@dataclass
class ExtensionConfig(AbstractConfig):
    """
    represents an extension configuration
    """

    enabled: Optional[bool]
    """True: The extension is enabled"""

    module: Optional[str]
    """Extension module name"""

    settings: Optional[Dict[str, Any]]
    """Settings specific to the extension"""

    def __post_init__(self):
        self.enabled = True if self.enabled is None else self.enabled


@dataclass
class Config(AbstractConfig):
    """
    represents a complete Encab configuration
    """

    encab: Optional[EncabConfig]
    """the basic encab configuraion"""

    extensions: Optional[Dict[str, ExtensionConfig]]
    """a dictionary with extension manes and their configuration"""

    programs: Optional[Dict[str, ProgramConfig]]
    """a dictionary with program names and their program configuration"""

    def __post_init__(self):
        if not self.encab:
            self.encab = EncabConfig.create()

    @staticmethod
    def load(stream: io.TextIOBase) -> "Config":
        """
        loads a configuration from a Yaml stream

        :param stream: the Yaml stream
        :type stream: io.TextIOBase
        :raises ConfigError: if the Yaml file is invalid or doesn't match configuration requirements
        :return: the configuration
        :rtype: Config
        """
        try:
            ConfigSchema = marshmallow_dataclass.class_schema(Config)
            config = ConfigSchema().load(yaml.safe_load(stream))
            assert isinstance(config, Config)
            return config
        except YAMLError as e:
            raise ConfigError(f"YAML error(s) {str(e)}")
        except ValidationError as e:
            msg = e.args[0]
            if isinstance(msg, dict):
                msg = yaml.dump(msg, default_flow_style=False)

            raise ConfigError(f"\n\n{msg}")
        except MarshmallowError as e:
            raise ConfigError(e.args)



import os
import pwd

from typing import Dict, Optional, Callable, Any, List, Union

from subprocess import Popen, PIPE
from signal import SIGKILL, SIGTERM

from logging import Logger, INFO, ERROR
from .log_stream import LogStream

from pwd import getpwnam
from grp import getgrnam

def getGroupId(name: str) -> int:
    """
    returns the GID for the group name

    :param name: the group name
    :type name: str
    :return: the GID
    :rtype: int
    :raise: KeyError if the group does not exist
    """
    return getgrnam(name).gr_gid

def getUserId(name: str) -> int:
    """
    returns the UID for the user name

    :param name: the user name
    :type name: str
    :return: the UID
    :rtype: int 
    :raise: KeyError if the User does not exist
    """
    return getpwnam(name).pw_uid


class Process(object):
    """
    Wrapper for POpen that supports logging of stdout and stdin
    """
    
    @staticmethod
    def update_current(user: Optional[int] = None, 
                       group: Optional[int] = None, 
                       umask: Optional[int] = None):
        """
        sets the current process user, group and umask

        :param user: the UID, defaults to None
        :type user: Optional[int], optional
        :param group: the GID, defaults to None
        :type group: Optional[int], optional
        :param umask: the mask, defaults to None
        :type umask: Optional[int], optional
        :raises ValueError: _description_
        """
        
        if group:
            os.setgid(group)
            
        if user and os.getuid() != user:
            try:
                user_data = pwd.getpwuid(user)
            except KeyError as e:
                raise ValueError(f"No passwd entry for user id: {user}", e)
                       
            os.initgroups(user_data.pw_name, user_data.pw_gid)
            os.setuid(user)

        if umask and umask != -1:
            os.umask(umask)

    
    def __init__(self, 
                 args:Union[str, List[str]], 
                 environment: Dict[str, str], 
                 user: Optional[int] = None, 
                 group: Optional[int] = None, 
                 umask: Optional[int] = None, 
                 shell: bool=False,
                 start_new_session: bool=True,
                 cwd: Optional[str] = None) -> None:
        self._args = args
        self._env = environment
        self._user = user
        self._group = group
        self._umask = umask
        self._shell = shell
        self._start_new_session = start_new_session
        self._process: Optional[Popen] = None
        self._cwd = cwd
        
    def execute(self, 
                exec: Callable[[Popen], Any], 
                stdin: Optional[int] = None, 
                stdout: Optional[int] = None, 
                stderr: Optional[int] = None) -> int:
        uid = self._user
        gid = self._group
        umask = self._umask

        if uid and os.getuid() != uid:
            try:
                user_data = pwd.getpwuid(uid)
            except KeyError as e:
                raise ValueError(f"No passwd entry for user id: {uid}", e)
            
        def preexec_fn():
            if gid:
                os.setgid(gid)
                
            if uid and user_data:
                os.initgroups(user_data.pw_name, user_data.pw_gid)
                os.setuid(uid)

            if umask and umask != -1:
                os.umask(umask)

        
        self._process = Popen(
                self._args,
                stdin=stdin,
                stdout=stdout,
                stderr=stderr,
                env=self._env,
                preexec_fn=preexec_fn,
                shell=self._shell,
                start_new_session=self._start_new_session,
                cwd=self._cwd
            )
        exec(self._process)
        self._process.wait()
        return self._process.returncode

    def execute_and_log(self, 
                exec: Callable[[Popen], Any], 
                logger: Logger,
                extra: Any, 
                log_stdout:bool=True) -> int:

        streams: List[LogStream] = list()

        def outer_exec(process: Popen) -> None:
            assert process.stdout
            assert process.stderr
            
            streams.append(LogStream(logger, ERROR, process.stderr, extra).start())
            
            if log_stdout:
                streams.append(LogStream(logger, INFO, process.stdout, extra).start())

            exec(process)

        try:
            return self.execute(outer_exec, None, PIPE, PIPE)
        finally:
            for stream in streams:
                stream.close()
    
    
    def pid(self) -> Optional[int]:
        return self._process.pid if self._process and self._process.pid else None
    
    def if_running(self, f: Callable[[int], Any]) -> Any:
        return f(self._process.pid) if self._process and self._process.pid else None
                    
    def signal(self, signal: int):
        try:
            self.if_running(lambda pid: os.kill(pid, signal))
        except ProcessLookupError:
            pass
    
    def kill(self):
        self.signal(SIGKILL)
        
    def terminate(self):
        self.signal(SIGTERM)

    
    


import sys
import os
import io

from logging import (
    getLogger,
    StreamHandler,
    Formatter,
)

from signal import SIGTERM, SIGINT, signal, getsignal
from typing import Optional, List, Tuple, Dict
from textwrap import shorten
from threading import Event

from .common.process import Process
from .config import Config, ConfigError
from .program_state import LoggingProgramObserver
from .program import ExecutionContext
from .programs import Programs
from .extensions import extensions, ENCAB

from .ext.log_sanitizer import LogSanitizerExtension
from .ext.validation import ValidationExtension
from .ext.startup_script import StarupScriptExtension


def load_config(encab_stream: Optional[io.TextIOBase] = None) -> Tuple[Config, str]:
    """
    Loads the configuration file

    :param encab_stream: _description_, defaults to None
    :type encab_stream: Optional[io.TextIOBase], optional
    :raises FileNotFoundError: _description_
    :return: _description_
    :rtype: Tuple[Config, str]
    """

    encab_file = None

    if encab_stream is not None:
        encab_file = "stream"

    source = "Argument"

    ENCAB_CONFIG = "ENCAB_CONFIG"

    if not encab_file:
        if ENCAB_CONFIG in os.environ:
            encab_file = os.environ[ENCAB_CONFIG]
            source = f"Environment {ENCAB_CONFIG}"

    ENCAB_FILE_CANDIDATES = [
        "./encab.yml",
        "./encab.yaml",
        "/etc/encab.yml",
        "/etc/encab.yaml",
    ]

    if not encab_file:
        for candidate in ENCAB_FILE_CANDIDATES:
            if os.path.exists(candidate):
                encab_file = candidate
                source = "Default location"

    if not encab_file:
        candidates = ", ".join(ENCAB_FILE_CANDIDATES)
        raise FileNotFoundError(f"Encab file not found in {candidates}.")

    if encab_stream:
        config = Config.load(encab_stream)
    else:
        with open(encab_file, "r") as f:
            config = Config.load(f)

    ENCAB_DRY_RUN = "ENCAB_DRY_RUN"

    dry_run = None
    if ENCAB_DRY_RUN in os.environ:
        value = os.environ[ENCAB_DRY_RUN]
        if not value:
            pass
        elif value == "1":
            dry_run = True
        elif value == "0":
            dry_run = False
        else:
            raise ConfigError(
                "Environment variable ENCAB_DRY_RUN"
                " expected to be '1' or '0' if set"
                f" but was '{value}'."
            )

    assert config.encab
    if dry_run is not None:
        config.encab.dry_run = dry_run

    return (config, f"file {encab_file}, source: {source}.")


def set_up_logger(config: Config):
    """
    Sets up the encab logger

    :param config: the encab configuraion
    :type config: Config
    :return: the logger
    :rtype: Logger
    """

    if config.encab:
        root_logger = getLogger()

        handler = StreamHandler()
        formatter = Formatter(config.encab.logformat)

        handler.setFormatter(formatter)

        loglevel = config.encab.loglevel
        assert isinstance(loglevel, int) or isinstance(loglevel, str)

        root_logger.setLevel(loglevel)
        root_logger.addHandler(handler)

    logger = getLogger(ENCAB)
    extensions.update_logger(ENCAB, logger)
    return logger


def set_up_extensions(config: Config, logger, extra: Dict[str, str]):
    """
    sets up and configures the extensions.
    In case of dry run, `encab.extensions.Extensions.validate_extension` instead of
    `encab.extensions.Extensions.configure_extension` is run for each plugin.

    :param config: the encab config
    :type config: Config
    :param logger: the logger to be used
    :type logger: Logger
    :param extra: the logger extension
    :type extra: Dict[str, str]
    """

    if config.extensions:
        assert config.encab and isinstance(config.encab.dry_run, bool)
        dry_run = config.encab.dry_run

        for name, econf in config.extensions.items():
            if econf.module:
                extensions.register_module(econf.module)

            assert isinstance(econf.enabled, bool)

            if dry_run:
                extensions.validate_extension(name, econf.enabled, econf.settings or {})
            else:
                extensions.configure_extension(
                    name, econf.enabled, econf.settings or {}
                )


def encab(
    encab_stream: Optional[io.TextIOBase] = None, args: Optional[List[str]] = None
):
    """
    Encab's main routine

    - loads the configuration file
    - replaces the main program with the one from arguments if specified
    - runs the programs

    :param encab_stream: the configruration as text stream, defaults to None
    :type encab_stream: Optional[io.TextIOBase], optional
    :param args: the main program arguments, defaults to None
    :type args: Optional[List[str]], optional
    """

    extensions.register(
        [StarupScriptExtension(), LogSanitizerExtension(), ValidationExtension()]
    )

    logger = None
    try:
        config, location = load_config(encab_stream)

        logger = set_up_logger(config)

        extra = {"program": ENCAB}

        logger.info("encab 0.0.9", extra=extra)
        logger.info("Using configuration %s", location, extra=extra)

        logger.debug(
            "Encab config: %s",
            shorten(str(config), width=127, placeholder="..."),
            extra=extra,
        )

        assert config.encab and isinstance(config.encab.dry_run, bool)
        dry_run = config.encab.dry_run

        if dry_run:
            logger.info("Dry run. No program will be started.", extra=extra)

        set_up_extensions(config, logger, extra)

        if dry_run:
            logger.info("Dry run succeeded. Exiting.", extra=extra)
            return

        config.encab.set_user()
        config.encab.set_group()

        assert config.encab.user is None or isinstance(config.encab.user, int)
        assert config.encab.group is None or isinstance(config.encab.group, int)
        assert config.encab.umask is None or isinstance(config.encab.umask, int)

        Process.update_current(
            config.encab.user, config.encab.group, config.encab.umask
        )

        program_config = config.programs or {}

        if program_config:
            logger.debug("Starting program(s)...", extra=extra)

        observer = LoggingProgramObserver(ENCAB, logger, extra)
        context = ExecutionContext(dict(os.environ), observer)

        if config.encab.environment:
            context = context.extend(config.encab.environment)

        args = args if args is not None else sys.argv[1:]

        programs = Programs(program_config, context, args, config.encab)

        def on_signal(signal, _):
            signames = {SIGINT: "SIGINT", SIGTERM: "SIGTERM"}

            logger.info(
                "Received %s. Interrupting/terminating programs...",
                signames[signal],
                extra=extra,
            )

            if signal == SIGINT:
                programs.interrupt()
                logger.info("Programs interrupted. Exiting.", extra=extra)
            else:
                programs.terminate()
                logger.info("Programs terminated. Exiting.", extra=extra)

            exit(0)

        sigint_handler = getsignal(SIGINT)
        sigterm_handler = getsignal(SIGTERM)

        signal(SIGINT, on_signal)
        signal(SIGTERM, on_signal)

        programs.run()

        if config.encab.halt_on_exit:
            logger.info("Programs ended. Halting for diagnose.", extra=extra)
            signal(SIGINT, sigint_handler)
            signal(SIGTERM, sigterm_handler)
            Event().wait()
        else:
            logger.debug("Programs ended. Exiting.", extra=extra)
    except PermissionError as e:
        print(
            f"Failed to set the encab user: {str(e)}."
            " \nTo set the user, you have to run encab as root."
        )
        exit(3)
    except IOError as e:
        print(f"I/O Error: {str(e)}")
        exit(1)
    except ValueError as e:
        print(f"Error in configuration: {str(e)}")
        exit(2)
    except KeyboardInterrupt:
        print("Encab was interrupted.")
        return


if __name__ == "__main__":
    encab()
